package PLPerl::CheckTrigger;

use strict;
use Carp;

our $VERSION = '1.0';

# Preloaded methods go here.
sub new
{
	my $class = shift;
	my $self = {};
	$self->{trigger} = shift;
	$self->{errmsg} = "";
	return bless $self, $class;
}

sub check_columns
{
	my $self = shift;
	my $options_arg = shift;
	my @column_names = @_;
	
	my @allowed_options = qw/ALLOW_LIST ONLY_FROM_NULL/;
	my $oldtuple = $self->{trigger}{old};
	my $newtuple = $self->{trigger}{new};
	my $tablename = $self->{trigger}{table_name};
	my @deny_columns;
	my %options;
   	
	
    # Treat the first argument as a list of options (which could be empty)
	%options = map {$_ => 0} @allowed_options;
	
    foreach my $option (split (/,\s*/, $options_arg)) {
		if(grep $_ eq $option, @allowed_options) { 
			# Note that this version of grep is supposedly faster
            $options{$option} = 1;
        } else {
            croak ("Unknown option ('$option') specified in first argument (allowed options are " . join(' ', @allowed_options) . ")");
		}
	}

	# check if all passed columns really exist 
	foreach (@column_names) {
		croak "column \"$_\" doesn't exist in relation $tablename" 
			unless exists $newtuple->{$_};
	}

	# Get the resulting set of columns. If ALLOW_LIST is true then we should
	# block only those columns not included in the list passed as a first param.

	if (!$options{ALLOW_LIST}) {
		@deny_columns = @column_names;
	} else {

		# Make a hash of column names to speed up searching
		my %column_names_hash;

		foreach my $colname (@column_names) {
			$column_names_hash{$colname} = 1;
		}

		# Get all relation columns that are not in column_names list and put
		# their names to the deny list.
		foreach my $colname (keys %{$oldtuple}) {
			push @deny_columns, $colname unless defined $column_names_hash{$colname};
		}
	}

	# Disallow updates for each column from deny list unless ONLY_FROM_NULL
	# is set and old tuple has NULL in the column to be updated
	foreach (@deny_columns) {
        if ($oldtuple->{$_} ne $newtuple->{$_}) {
        	unless ($options{ONLY_FROM_NULL} == 1 and !defined $oldtuple->{$_}) {
				$self->{errmsg} = sprintf "update of attribute \"%s\" denied by the trigger \"%s\"", $_, $self->{trigger}{name};
				return 0;
			}
        }
	}
	return 1;
}

sub check_condition 
{
	my $self = shift;
	my $condition = shift;
	my $return_condition = shift;
	my @columns = @_;
	my @args;
	my $argno;
	my $nargs;
	
	::elog(::DEBUG, "condition: $condition");
	# prepare a list of arguments, substituting OLD and NEW with perl equivalents
	foreach (@columns) {
		if (/^(OLD|NEW)\.(.*)$/) {
			my ($tupname, $attname) = ($1, $2);
			
			# double backslashes in $2
			$attname =~ s/\\/\\\\/;
			# also escape single quote characters.
			$attname =~ s/\'/\\\'/;
		 	my $argname = '$self->{trigger}{'.lc($tupname).'}{\''.$attname.'\'}';
			::elog(::DEBUG, "substition result: $argname");
			
			my $val = eval $argname;
			
			# do additional processing of non-numeric values
			if ((defined $val) && ($val =~ /\D/)) {
				# double backslashes in argument
				$val =~ s/\\/\\\\/;
				# strip first and last quotes
				$val =~s/^\'(.*)\'$/$1/;
				# escape single quote characters
				$val =~ s/\'/\\\'/;
				# put value in quotes unless it is already quoted
				$val = "'".$val."'" unless $val =~ /^\'.*\'$/;
				# treat the argument as escape string
				$val = 'E'.$val;
			} elsif (!defined $val) {
				$val = 'NULL';
			}
			
			::elog(::DEBUG, "argument value: $val");
			push @args, $val;
		} else {
			croak "argument $argno should be the reference to a column of OLD or NEW tuples";
		}
		$argno++;
	}
	
	# form a condition string to check
	# check that the number of actual arguments matches the number of expected.
	$nargs = () = $condition =~ /%s/g;
	carp "Wrong number of arguments in sprintf: expected $nargs, passed ".scalar(@args) 
	unless (scalar(@args) == $nargs);
	
	$condition = sprintf $condition, @args;
	my $condition_sql = "SELECT $condition AS result";
	::elog(::DEBUG, "condition_sql: $condition_sql, condition: $condition");
	
	# execute condition as SQL and get the result
	return $condition_sql if ($return_condition);
	my $res = ::spi_exec_query($condition_sql);
	my $result = $res->{rows}[0]->{result};
	
	croak "expression $condition_sql should return exactly one row" unless $res->{processed} == 1;
	croak "expression $condition_sql result is not defined" unless defined $result;
	
	::elog(::DEBUG, "result: $result");
	
	if ($result eq 't') {
		::elog(::DEBUG, "allow");
		return 1;
	} elsif ($result eq 'f') {
		::elog(::DEBUG, "deny");
		$self->{errmsg} = sprintf "expression %s is false, %s is not allowed", 
			$condition, $self->{trigger}{event};
		return 0;
	}
	
	croak "expression $condition_sql should return a result of boolean type";	
}

# quote nullable
sub pg_quote_nullable {
	my $str = shift;
	if ($str ne '') {
		for my $var ('a'..'zzz') {
			if ($str !~ /\$_${var}_\$/) {
				return '$_'.$var.'_$'."$str".'$_'.$var.'_$' ;
			}
		}
		carp "unexpected string: $str";
		return $str;	
	}
	return 'NULL'; 
}

1;
__END__

=head1 NAME

PLPerl::CheckTrigger - a helper module for check_updates postgres triggers

=head1 SYNOPSIS

  use PLPerl::CheckTrigger;

=head1 DESCRIPTION

This module is designed to be called from a PostgreSQL trigger function
written in PL/Perlu. It provides 2 methods: check_columns and check_condition.
These methods should be called from the row on update trigger. Their return
value is 1 if the underlying operation (update) is allowed and 0 if it is denied.

check_columns gets a column list, an options list and old/new tuples for the
relation. If options list has ALLOW_LIST option set, then only updates for
columns from the specified list are allowed, otherwise updates for those and only
those columns are denied. 

check_condition gets a condition string and a list of columns and evaluates
the condition using column values. Column names should be specified in a form
(OLD|NEW).column, i.e. NEW.name or OLD.val. Condition string uses % as a
placeholder for column names.


=head1 AUTHOR

Alexey Klyukin, E<lt>alexk@commandprompt.comE<gt>. Development is
sponsored by Enovafinancial (www.enovafinancial.com).

=head1 COPYRIGHT AND LICENSE

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.


=cut
